/*
 * Extra Windows Position Manager for Pidgin (Mostly MS Windows)
 *
 * Pidgin is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#define  _CRT_SECURE_NO_WARNINGS
#include <windows.h>
#include <windowsx.h>
#include <assert.h>

#include "prefs.h"
#include "version.h"
#include "gtkblist.h"
#include "gtkplugin.h"
#include "gtkconv.h"
#include "gtkprefs.h"
#include "gtkutils.h"
#include "debug.h"

#include "extpos.h"

extpos_data_t g_extpos_data = {0};

// Pidgin window gtk_roles
static const char *s_roles[] =
{
    "conversation",
    "buddy_pounce",
    "pounces",
    "log_viewer",
    "add_buddy",
    "assoc_buddy",
    "new_person",
    "ticker",
    "account",
    "accounts",
    "join_chat",
    "buddy_list",
    "add_chat",
    "debug",
    "about",
    "preferences",
    "file transfer",
    "smiley_dialog",
    "notify_dialog",
    "new_mail_detailed",
    "plugin_config",
    "plugins",
    "privacy",
    "multifield",
    "room list",
    "status",
    "statuses",
    "substatus",
};

enum
{
    COLUMN_ROLE,
    COLUMN_SET_POSITION,
    COLUMN_X,
    COLUMN_Y,
    COLUMN_W,
    COLUMN_H,
    COLUMN_NUM
};

enum
{
    POS_X,
    POS_Y,
    POS_W,
    POS_H,
    POS_Enabled,
    POS_Max
};

static void
dprintf(const char *szFormat, ...)
{
    char buffer[1024];

    va_list arglist;
    va_start(arglist, szFormat);

    vprintf(szFormat, arglist);

    // Write also to the debug output spew
    g_vsnprintf(buffer, sizeof(buffer) - 1, szFormat, arglist);
    OutputDebugString( buffer );

    va_end(arglist);
}

//=========================================================================
// Given a native win32 hwnd, try to find the role Pidgin has associated
//  with it. Should be something like "conversation", "buddy_list", etc.
//  Grep the Pidgin source for gtk_window_set_role(...) for others.
//=========================================================================
const gchar *
toplevel_window_get_gtk_role(HWND hwnd)
{
    GList *l;
    G_CONST_RETURN gchar *role = NULL;
    GdkWindow *gdkwin = gdk_window_lookup(*(GdkNativeWindow *)&hwnd);

    for(l = gtk_window_list_toplevels(); l != NULL; l = l->next)
    {
        GtkWidget *widget = GTK_WIDGET(l->data);

        if(gdkwin == widget->window)
        {
            GtkWindow *win = GTK_WINDOW(widget);
            role = gtk_window_get_role(win);
            break;
        }
    }

    return role ? role : "--role_unknown--";
}

//=========================================================================
// Get the Window title replacing spaces and '/' characters.
//=========================================================================
static void
get_window_title_safe(HWND hwnd, gchar *title, guint max_title_size)
{
    wchar_t utf16_title[512];

    // Return nil string if something fails below.
    title[0] = 0;

    // Get the unicode title.
    if(GetWindowTextW(hwnd, utf16_title, SIZEOF_ARRAY(utf16_title)))
    {
        char *utf8_title;

        // Convert unicode to utf8 string.
        utf8_title = g_utf16_to_utf8(utf16_title, -1, NULL, NULL, NULL);
        if(utf8_title)
        {
            strncpy(title, utf8_title, max_title_size);
            title[max_title_size - 1] = 0;

            g_free(utf8_title);
        }
    }
}

//=========================================================================
// Given a string with comma delimited numbers, return their values in an
//  array.
//=========================================================================
static int
get_comma_delimited_values(const gchar *values, int *nums, int nums_max)
{
    int i = 0;

    if(values && values[0])
    {
        for(; i < nums_max; i++)
        {
            char *endptr;

            nums[i] = strtoul(values, &endptr, 10);

            values = strchr(values, ',');
            if(!values)
                break;
            values++;
        }
    }

    return i;
}

//=========================================================================
// Format the string and return the prefs XML key string.
//=========================================================================
static gchar *
get_prefs_pos_key(gchar *pref, guint pref_size, HWND hwnd, gchar const *role)
{
    gchar title[128];

    // Get the window title.
    get_window_title_safe(hwnd, title, sizeof(title));

    g_snprintf(pref, pref_size, "/plugins/gtk/extpos/positions/%s::%s", role, title);
    return pref;
}

//=========================================================================
// Format the string and return the prefs XML key string.
//=========================================================================
static gchar *
get_prefs_role_key(gchar *pref, guint pref_size, gchar const *role)
{
    g_snprintf(pref, pref_size, "/plugins/gtk/extpos/roles/%s", role);
    return pref;
}

//=========================================================================
// Format the string and return the forcesize XML key string.
//=========================================================================
static gchar *
get_forcesize_role_key(gchar *pref, guint pref_size, gchar const *role)
{
    g_snprintf(pref, pref_size, "/plugins/gtk/extpos/forcesize/%s", role);
    return pref;
}

//=========================================================================
// varargs version of purple_prefs_set_string.
//=========================================================================
static void
purple_prefs_set_string_v(gchar const *prefs_pos_key, gchar const *format, ...)
{
    va_list args;
    gchar value[256];

    va_start(args, format);
    g_vsnprintf(value, sizeof(value), format, args);
    va_end(args);

    purple_prefs_set_string(prefs_pos_key, value);
}

//=========================================================================
// varargs version of purple_prefs_add_string.
//=========================================================================
static void
purple_prefs_add_string_v(gchar const *prefs_pos_key, gchar const *format, ...)
{
    va_list args;
    gchar value[256];

    va_start(args, format);
    g_vsnprintf(value, sizeof(value), format, args);
    va_end(args);

    purple_prefs_add_string(prefs_pos_key, value);
}

//=========================================================================
// Check if the window is iconified or maximized.
//=========================================================================
static gboolean
is_window_iconified_or_maximized(HWND hwnd)
{
    return IsMaximized(hwnd) || IsMinimized(hwnd);
}

//=========================================================================
// Update the x, y, w, h information for a disabled role.
//=========================================================================
static void
update_role_xywh(const gchar *role, RECT *rc)
{
    int i;
    gchar path[10];
    GtkTreeIter iter;
    LONG xywh[4];

    assert(g_extpos_data.model);

    // Find the path string to our role.
    path[0] = 0;
    for(i = 0; i < sizeof(s_roles) / sizeof(s_roles[0]); i++)
    {
        if(!strcmp(s_roles[i], role))
        {
            g_snprintf(path, sizeof(path), "%d", i);
            break;
        }
    }
    if(!path[0])
        return;

    g_return_if_fail(gtk_tree_model_get_iter_from_string(
        GTK_TREE_MODEL(g_extpos_data.model), &iter, path));

#ifdef _DEBUG
    {
        // We should only be updating role info that aren't enabled.
        gboolean enabled;
        gtk_tree_model_get(GTK_TREE_MODEL(g_extpos_data.model), &iter,
                           COLUMN_SET_POSITION, &enabled,
                           -1);
        assert(!enabled);
    }
#endif

    xywh[0] = rc->left;
    xywh[1] = rc->top;
    xywh[2] = rc->right - rc->left;
    xywh[3] = rc->bottom - rc->top;

    // Go through and set the x, y, w, h data.
    for(i = 0; i < 4; i++)
    {
        GValue val;

        val.g_type = 0;
        gtk_tree_model_get_value(GTK_TREE_MODEL(g_extpos_data.model),
            &iter, i + COLUMN_X, &val);

        if(xywh[i] != g_value_get_int(&val))
        {
            gtk_list_store_set(g_extpos_data.model, &iter,
                i + COLUMN_X, xywh[i], -1);
        }
        g_value_unset(&val);
    }
}

//=========================================================================
// Move/Size window callback. Stores the latest position.
//=========================================================================
static void
hwnd_movesized(HWND hwnd, RECT *rc)
{
    const gchar *values;
    gchar prefs_key[256];
    int pos[POS_Max] = {0};
    const gchar *role = toplevel_window_get_gtk_role(hwnd);

    // See if the role key is there and enabled.
    get_prefs_role_key(prefs_key, sizeof(prefs_key), role);
    values = purple_prefs_get_string(prefs_key);
    get_comma_delimited_values(values, pos, sizeof(pos) / sizeof(pos[0]));

    // Only store the data for this window if the role data isn't enabled.
    if(!pos[POS_Enabled])
    {
        // If this role isn't updated, then update the information if
        //  we have a gtk tree model.
        if(g_extpos_data.model)
            update_role_xywh(role, rc);

        // Get the prefs location string.
        get_prefs_pos_key(prefs_key, sizeof(prefs_key), hwnd, role);

        // Store the new location for this window.
        purple_prefs_set_string_v(prefs_key, "%d,%d,%d,%d",
            rc->left, rc->top, rc->right - rc->left, rc->bottom - rc->top);
    }
}

//=========================================================================
// Callback for whenever the conversation/[im|chat]/default_width pref is
//  changed. This was the only way I could find to get called when a new window
//  was brought up. Hopefully it'll continue to work.
//=========================================================================
static void
hwnd_created(HWND hwnd)
{
    const gchar *role;
    const gchar *values;
    char prefs_key[256];
    int pos[POS_Max] = {0};
    UINT flags = SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_NOACTIVATE;

    if(is_window_iconified_or_maximized(hwnd))
        return;

    role = toplevel_window_get_gtk_role(hwnd);

    // Try and get the role data
    get_prefs_role_key(prefs_key, sizeof(prefs_key), role);
    values = purple_prefs_get_string(prefs_key);
    get_comma_delimited_values(values, pos, sizeof(pos) / sizeof(pos[0]));

    // If the role data isn't enabled, then do the per window title thing.
    if(!pos[POS_Enabled])
    {
        int default_x, default_y;

        memset(pos, 0, sizeof(pos));

        // Grab the default x,y position.
        default_x = purple_prefs_get_int("/plugins/gtk/extpos/default_x");
        default_y = purple_prefs_get_int("/plugins/gtk/extpos/default_y");

        // Get the prefs location string.
        get_prefs_pos_key(prefs_key, sizeof(prefs_key), hwnd, role);

        // Add the pref for this window at the default x,y.
        purple_prefs_add_string_v(prefs_key, "%d,%d", default_x, default_y);

        // Grab the x,y,w,h values
        values = purple_prefs_get_string(prefs_key);
        get_comma_delimited_values(values, pos, sizeof(pos) / sizeof(pos[0]));

        // If people aren't storing the width and height, then clear
        //  it out.
        if(!purple_prefs_get_bool("/plugins/gtk/extpos/restore_win_sizes"))
            pos[POS_W] = pos[POS_H] = 0;
    }

    // See if there is an entry under the forcesize key for this role. If so,
    //  then we will replace the width and height values with whatever they're
    //  telling us to use.
    get_forcesize_role_key(prefs_key, sizeof(prefs_key), role);
    values = purple_prefs_get_string(prefs_key);
    get_comma_delimited_values(values, &pos[POS_W], 2);

    if(pos[POS_W] <= 0 || pos[POS_H] <= 0)
        flags |= SWP_NOSIZE;

    SetWindowPos(hwnd, NULL, pos[POS_X], pos[POS_Y], pos[POS_W], pos[POS_H], flags);
}

//=========================================================================
// Toggle the buddy list visibility.
//=========================================================================
static void
hotkey_callback_toggle_blist(extpos_data_t *extpos_data)
{
    pidgin_blist_toggle_visibility();
}

//=========================================================================
// Cycle through the conversation windows.
//=========================================================================
static void
hotkey_callback_cycle_convs(extpos_data_t *extpos_data)
{
    GList *list = NULL;

    if(extpos_data->last_conv_list)
    {
        for(list = pidgin_conv_windows_get_list(); list; list = list->next)
        {
            if(extpos_data->last_conv_list == list)
            {
                list = list->next;
                break;
            }
        }
    }

    if(!list)
        list = pidgin_conv_windows_get_list();

    if(list)
    {
        PidginWindow *win = list->data;

        gtk_window_present(GTK_WINDOW(win->window));
        extpos_data->last_conv_list = list;
    }
}

//=========================================================================
// Standard plugin load stuff
//=========================================================================
static gboolean
plugin_load(PurplePlugin *plugin)
{
    if(g_extpos_data.is_loaded)
    {
        purple_debug_error("extpgos", "plugin_load with plugin already loaded?");
        return FALSE;
    }
    g_extpos_data.is_loaded = 1;

    extpos_register_hotkey(&g_extpos_data,
        IDHK_TOGGLE_BLIST,
        purple_prefs_get_int("/plugins/gtk/extpos/hotkey_toggle_blist_mods"),
        purple_prefs_get_int("/plugins/gtk/extpos/hotkey_toggle_blist_vk"),
        hotkey_callback_toggle_blist);

    extpos_register_hotkey(&g_extpos_data,
        IDHK_CYCLE_CONVS,
        purple_prefs_get_int("/plugins/gtk/extpos/hotkey_cycle_convs_mods"),
        purple_prefs_get_int("/plugins/gtk/extpos/hotkey_cycle_convs_vk"),
        hotkey_callback_cycle_convs);

    return hook_init(&g_extpos_data, hwnd_created, hwnd_movesized);
}

//=========================================================================
// Standard plugin unload stuff
//=========================================================================
static gboolean
plugin_unload(PurplePlugin *plugin)
{
    if(!g_extpos_data.is_loaded)
    {
        purple_debug_error("extpgos", "plugin_unload with plugin not loaded yet?");
        return FALSE;
    }
    g_extpos_data.is_loaded = 0;

    extpos_close_hotkeys(&g_extpos_data);

    hook_close(&g_extpos_data);

    gtk_list_store_clear(g_extpos_data.model);

    memset(&g_extpos_data, 0, sizeof(g_extpos_data));

    return TRUE;
}

//=========================================================================
// Store the role data for whatever line the treeiter is on.
//=========================================================================
static void
save_role_data(GtkTreeIter *iter, int enabled)
{
    int i;
    int pos[4];
    GValue val_role;
    char role_key[256];

    // If they didn't have the enabled information, then snag it.
    if(enabled == -1)
    {
        gtk_tree_model_get(GTK_TREE_MODEL(g_extpos_data.model), iter,
                           COLUMN_SET_POSITION, &enabled,
                           -1);
    }

    // Go through and get x,y,w,h.
    for(i = 0; i < sizeof(pos) / sizeof(pos[0]); i++)
    {
        GValue val;

        val.g_type = 0;
        gtk_tree_model_get_value(GTK_TREE_MODEL(g_extpos_data.model), iter,
            i + COLUMN_X, &val);
        pos[i] = g_value_get_int(&val);

        g_value_unset(&val);
    }

    // Get the role key.
    val_role.g_type = 0;
    gtk_tree_model_get_value(GTK_TREE_MODEL(g_extpos_data.model), iter,
        COLUMN_ROLE, &val_role);
    get_prefs_role_key(role_key, sizeof(role_key), g_value_get_string(&val_role));

    // Store POS_X, POS_Y, POS_W, POS_H, POS_Enabled.
    purple_prefs_set_string_v(role_key, "%d,%d,%d,%d,%d",
        pos[0], pos[1], pos[2], pos[3], enabled);

    g_value_unset(&val_role);
}

//=========================================================================
// The w, y, w, or h data for a role was modified. Update prefs.xml.
//=========================================================================
static void
xywh_edited(GtkCellRendererText *cellrenderertext, gchar *path,
    gchar *arg2, gpointer data)
{
    int num;
    GValue val;
    GtkTreeIter iter;

    if(arg2[0] == '\0')
    {
        gdk_beep();
        return;
    }

    g_return_if_fail(gtk_tree_model_get_iter_from_string(
        GTK_TREE_MODEL(g_extpos_data.model), &iter, path));

    val.g_type = 0;
    gtk_tree_model_get_value(GTK_TREE_MODEL(g_extpos_data.model), &iter,
        GPOINTER_TO_INT(data), &val);

    num = atoi(arg2);
    if(num != g_value_get_int(&val))
    {
        gtk_list_store_set(g_extpos_data.model, &iter,
            GPOINTER_TO_INT(data), num, -1);

        save_role_data(&iter, -1);
    }
    g_value_unset(&val);
}

//=========================================================================
// They toggled the Set Position checkbox - store the new value.
//=========================================================================
static void
set_position_toggled(GtkCellRendererToggle *cellrenderertoggle, gchar *path,
    gpointer data)
{
    gboolean enabled;
    GtkTreeIter iter;

    g_return_if_fail(gtk_tree_model_get_iter_from_string(
        GTK_TREE_MODEL(g_extpos_data.model), &iter, path));

    gtk_tree_model_get(GTK_TREE_MODEL(g_extpos_data.model), &iter,
                       COLUMN_SET_POSITION, &enabled,
                       -1);

    enabled = !enabled;
    gtk_list_store_set(GTK_LIST_STORE(g_extpos_data.model), &iter,
                       COLUMN_SET_POSITION, enabled,
                       -1);

    save_role_data(&iter, enabled);
}

//=========================================================================
// Read in the role data from prefs.xml and init our gtk list store model.
//=========================================================================
static void
init_gtk_list_store_model()
{
    int i;
    int default_x;
    int default_y;

    if(g_extpos_data.model)
        return;

    default_x = purple_prefs_get_int("/plugins/gtk/extpos/default_x");
    default_y = purple_prefs_get_int("/plugins/gtk/extpos/default_y");

    g_extpos_data.model = gtk_list_store_new(
        (gint)COLUMN_NUM,
        G_TYPE_STRING,      // COLUMN_ROLE
        G_TYPE_BOOLEAN,     // COLUMN_SET_POSITION
        G_TYPE_INT,         // COLUMN_X
        G_TYPE_INT,         // COLUMN_Y
        G_TYPE_INT,         // COLUMN_W
        G_TYPE_INT);        // COLUMN_H

    for(i = 0; i < sizeof(s_roles) / sizeof(s_roles[0]); i++)
    {
        int pos[POS_Max];
        GtkTreeIter iter;
        char role_key[256];
        const gchar *values;

        get_prefs_role_key(role_key, sizeof(role_key), s_roles[i]);
        values = purple_prefs_get_string(role_key);
        if(!values)
        {
            pos[POS_Enabled] = FALSE;
            pos[POS_X] = default_x;
            pos[POS_Y] = default_y;
            pos[POS_W] = 0;
            pos[POS_H] = 0;
        }
        else
        {
            get_comma_delimited_values(values, pos,
                sizeof(pos) / sizeof(pos[0]));
        }

        gtk_list_store_append(g_extpos_data.model, &iter);
        gtk_list_store_set(g_extpos_data.model, &iter,
            COLUMN_ROLE, s_roles[i],
            COLUMN_SET_POSITION, pos[POS_Enabled],
            COLUMN_X, pos[POS_X],
            COLUMN_Y, pos[POS_Y],
            COLUMN_W, pos[POS_W],
            COLUMN_H, pos[POS_H],
            -1);
    }
}

//=========================================================================
// Preferences UI dialog thing
//=========================================================================
static GtkWidget *
get_config_frame(PurplePlugin *plugin)
{
    int i;
    GtkSizeGroup *sg;
    GtkWidget *ret = NULL;
    GtkWidget *frame = NULL;
    GtkWidget *vbox = NULL;

    GtkWidget *win;
    GtkWidget *tree;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    static const char *s_titles[] = { "Left", "Top", "Width", "Height" };

    // Init our gtk list store model tree with window role information
    init_gtk_list_store_model();

    ret = gtk_vbox_new(FALSE, 18);
    gtk_container_set_border_width(GTK_CONTAINER (ret), 12);
    sg = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    /*---------- Options ----------*/
    frame = pidgin_make_frame(ret, "Options");
    vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_add(GTK_CONTAINER(frame), vbox);

    // X / Left
    pidgin_prefs_labeled_spin_button(vbox, "Initial Window Position Left: ",
            "/plugins/gtk/extpos/default_x", -64000, 64000, sg);

    // Y / Top
    pidgin_prefs_labeled_spin_button(vbox, "Initial Window Position Top: ",
            "/plugins/gtk/extpos/default_y", -64000, 64000, sg);

    // Restore width/height
    pidgin_prefs_checkbox("Restore _Window Sizes",
        "/plugins/gtk/extpos/restore_win_sizes", vbox);

    // Formating toolbar. This so shouldn't be in this plugin but someone
    //  asked for it and it doesn't appear to be anywhere else, so here it is.
    pidgin_prefs_checkbox("_Show Conversation Formatting Toolbar",
        "/pidgin/conversations/show_formatting_toolbar", vbox);

    /*---------- Window Type Position Overrides tree ----------*/
    vbox = pidgin_make_frame(ret, "Window Type Position Overrides");
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_widget_show(vbox);

    win = gtk_scrolled_window_new(0, 0);
    gtk_box_pack_start(GTK_BOX(vbox), win, TRUE, TRUE, 0);
    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(win),
        GTK_SHADOW_IN);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
        GTK_POLICY_NEVER,       // hscrollbar_policy
        GTK_POLICY_ALWAYS);     // vscrollbar_policy
    gtk_widget_show(win);

    tree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(g_extpos_data.model));
    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree), TRUE);
    gtk_widget_set_size_request(tree, -1, 200);

    // Window type
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes(
        "Window Type", renderer,
        "text", COLUMN_ROLE,
        NULL);
    gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
    gtk_tree_view_column_set_fixed_width(column, 100);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

    // Set Position
    renderer = gtk_cell_renderer_toggle_new();
    g_object_set(G_OBJECT(renderer),
        "activatable", TRUE,
        NULL);
    g_signal_connect(G_OBJECT(renderer), "toggled",
        G_CALLBACK(set_position_toggled), NULL);
    column = gtk_tree_view_column_new_with_attributes(
        "Set Position", renderer,
        "active", COLUMN_SET_POSITION,
        NULL);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

    for(i = COLUMN_X; i <= COLUMN_H; i++)
    {
        // x, y, w, h
        renderer = gtk_cell_renderer_text_new();
        g_object_set(G_OBJECT(renderer), "editable", TRUE, NULL);
        g_signal_connect(G_OBJECT(renderer), "edited",
            G_CALLBACK(xywh_edited), GINT_TO_POINTER(i));
        column = gtk_tree_view_column_new_with_attributes(
            s_titles[i - COLUMN_X], renderer,
            "text", i,
            NULL);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
        gtk_tree_view_column_set_fixed_width(column, 50);
        gtk_tree_view_column_set_resizable(column, TRUE);
        gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
    }

    gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(tree)),
         GTK_SELECTION_MULTIPLE);
    gtk_container_add(GTK_CONTAINER(win), tree);
    gtk_widget_show(tree);

    gtk_widget_show_all(ret);
    return ret;
}

static PidginPluginUiInfo ui_info =
{
    get_config_frame,
    0 /* page_num (Reserved) */
};

#define PURPLE_WEBSITE      "http://pidgin.im/"
#define VERSION             "2.0.0"

static PurplePluginInfo info =
{
    PURPLE_PLUGIN_MAGIC,
    PURPLE_MAJOR_VERSION,
    PURPLE_MINOR_VERSION,
    PURPLE_PLUGIN_STANDARD,                         /**< type           */
    PIDGIN_PLUGIN_TYPE,                             /**< ui_requirement */
    0,                                              /**< flags          */
    NULL,                                           /**< dependencies   */
    PURPLE_PRIORITY_DEFAULT,                        /**< priority       */

    "gtk-extpos",                                   /**< id             */
    "ExtPos",                                       /**< name           */
    VERSION,                                        /**< version        */
    "Extra Windows Position Manager.",              /**< summary        */
    "Windows Position/Size Manager. "
        "(Built "  __DATE__ ")",                    /**  description    */
    "Michael Sartain <mikesart@gmail.com>",         /**< author         */
    PURPLE_WEBSITE,                                 /**< homepage       */

    plugin_load,                                    /**< load           */
    plugin_unload,                                  /**< unload         */
    NULL,                                           /**< destroy        */

    &ui_info,                                       /**< ui_info        */
    NULL,                                           /**< extra_info     */
    NULL,                                           /**< prefs_info     */
    NULL                                            /**< actions        */
};

static void
init_plugin(PurplePlugin *plugin)
{
    purple_prefs_add_none("/plugins/gtk/extpos");
    purple_prefs_add_none("/plugins/gtk/extpos/positions");
    purple_prefs_add_none("/plugins/gtk/extpos/roles");

    purple_prefs_add_int( "/plugins/gtk/extpos/default_x", 200);
    purple_prefs_add_int( "/plugins/gtk/extpos/default_y", 200);
    purple_prefs_add_bool("/plugins/gtk/extpos/restore_win_sizes", TRUE);

    // 1. Exit Pidgin.
    // 2. Edit prefs.xmls
    // 3. Set mods to something like: 8 (MOD_WIN)
    // 4. Set vk to something like: 71 ('G')
    // 5. Start Pidgin and Win+G should now cycle through your open conversation windows
    purple_prefs_add_int( "/plugins/gtk/extpos/hotkey_cycle_convs_mods", -1);
    purple_prefs_add_int( "/plugins/gtk/extpos/hotkey_cycle_convs_vk", -1);

    // Same as above with different hotkey will toggle blist visibility.
    purple_prefs_add_int( "/plugins/gtk/extpos/hotkey_toggle_blist_mods", -1);
    purple_prefs_add_int( "/plugins/gtk/extpos/hotkey_toggle_blist_vk", -1);
}

PURPLE_INIT_PLUGIN(extpos, init_plugin, info)

